'use strict';

const {createBluetooth} = require('node-ble');
let spawn = require('child_process').spawn;

let utl;
let dbLv;
let DBG_NAME = '';
const DEBG_API = 3;
const DEBG_ERR = 1;

///////////////////////////////////////////////////////////////

// Constructor
let NodeBleLib = function(utlFunc, dbgLevel, name, subName) {
	utl = utlFunc;
	dbLv = dbgLevel;
	DBG_NAME = name + " node-ble";
};

/*****************************************************************************
 * node-bleの開始
 *  timeout : TimeOut時間
 ****************************************************************************/
NodeBleLib.prototype.startLib = function(timeout) {
	return new Promise(async (resolve) => {
		let timId = setTimeout(() => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]startLib(new)=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]startLib(new)=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _startLib_sub(timId));
	});
};
function _startLib_sub(timId){
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "startLib開始(new)");
			const {bluetooth, destroy} = createBluetooth();
			const adapter = await bluetooth.defaultAdapter();
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "startLib完了(new)");

			clearTimeout(timId);
			ret.result  = true;
			ret.adapter = adapter;
			ret.destroy = destroy;
			ret.bluetooth = bluetooth;
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]startLib(new)=" + e);

			clearTimeout(timId);
			ret.result  = false;
			ret.adapter = null;
			ret.destroy = null;
			ret.bluetooth = null;
			ret.e = e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * node-bleのSession廃棄
 *  adapter      : bluetooth.defaultAdapter()
 *  destroy      : createBluetooth()
 *  timeout      : TimeOut時間
 ****************************************************************************/
NodeBleLib.prototype.destroy = function(adapter, destroy, timeout) {
	return new Promise(async (resolve) => {
		let timId = setTimeout(() => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]destroy()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]startLib(new)=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _destroy_sub(adapter, destroy, timId));
	});
};
function _destroy_sub(adapter, destroy, timId){
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			if(! await adapter.isDiscovering()){
				utl.debugLog(DEBG_API, dbLv, DBG_NAME, "destroy開始");
			}else{
				utl.debugLog(DEBG_API, dbLv, DBG_NAME, "discovery中のdestroy開始");
			}
			await destroy();

			clearTimeout(timId);
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "destroy完了");
			ret.result  = true;
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]destroy()=" + e);

			clearTimeout(timId);
			ret.result  = false;
			ret.e = e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * node-bleのPoweredの取得
 *  adapter      : bluetooth.defaultAdapter()
 *  timeout      : TimeOut時間
 ****************************************************************************/
NodeBleLib.prototype.isPowered = function(adapter, timeout) {
	return new Promise(async (resolve) => {
		let timId = setTimeout(() => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]isPowered()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]isPowered(new)=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _isPowered_sub(adapter, timId));
	});
};
function _isPowered_sub(adapter, timId){
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "isPowered開始");
			ret.isPowered = await adapter.isPowered();
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "isPowered完了" + ret.isPowered);

			clearTimeout(timId);
			ret.result  = true;
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]isPowered()=" + e);

			clearTimeout(timId);
			ret.result  = false;
			ret.e = e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * defaultAdapterのPower設定
 *  powered      : true/false
 *  timeout      : TimeOut時間
 ****************************************************************************/
NodeBleLib.prototype.adapterPower = function(powered, timeout) {
	return new Promise(async (resolve) => {
		let timId = setTimeout(() => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]adapterPower()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]adapterPower()=timeout";
			resolve(ret);
		}, timeout);
			let pwStr = "";
			if(powered){
				pwStr = "on";
			}else{
				pwStr = "off";
			}
		resolve(await _adapterPower_sub(pwStr, timId));
	});
};
function _adapterPower_sub(powerStr, timId) {
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			// bleIdはxx:xx:xx:xx:xx:xx形式
			let cmdExe = spawn('bluetoothctl', ['power', powerStr]);
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "adapterPower[" + powerStr + "]開始");
			cmdExe.stdout.setEncoding('utf8');

			cmdExe.stderr.once('data', (err) => {
				let resString = err.toString();
				utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]adapterPower[" + powerStr + "]=" + resString);
				clearTimeout(timId);

				ret.result = false;
				ret.e = resString;
				resolve(ret);
			});
			cmdExe.stdout.once('data', async (data) => {
				let resString = data.toString();			// Successful
				utl.debugLog(DEBG_API, dbLv, DBG_NAME, "adapterPower[" + powerStr + "]完了=" + resString);
				clearTimeout(timId);

				ret.result = true;
				resolve(ret);
			});
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]adapterPower()=" + e);
			ret.result = false;
			clearTimeout(timId);

			ret.e = e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * node-bleの探索開始
 *  adapter      : bluetooth.defaultAdapter()
 *  timeout      : API Timeout value.
 ****************************************************************************/
NodeBleLib.prototype.startDiscovery = function(adapter, timeout){
	return new Promise(async (resolve) => {
		let timId = setTimeout(() => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]startDiscovery()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]startDiscovery()=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _startDiscovery_sub(adapter, timId));
	});
};
function _startDiscovery_sub(adapter, timId){
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			if(! await adapter.isDiscovering()){
				utl.debugLog(DEBG_API, dbLv, DBG_NAME, "startDiscovery開始");
				await adapter.startDiscovery();
				utl.debugLog(DEBG_API, dbLv, DBG_NAME, "startDiscovery完了");
			}else{
				utl.debugLog(DEBG_API, dbLv, DBG_NAME, "既にdiscovery中");
				utl.debugLog(DEBG_API, dbLv, DBG_NAME, "startDiscovery完了");
			}

			clearTimeout(timId);
			ret.result = true;
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]startDiscovery()=" + e);

			clearTimeout(timId);
			ret.result = false;
			ret.e = e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * node-bleの探索停止
 *  adapter      : bluetooth.defaultAdapter()
 *  timeout      : API Timeout value.
 ****************************************************************************/
NodeBleLib.prototype.stopDiscovery = function(adapter, timeout) {
	return new Promise(async (resolve) => {
		let timId = setTimeout(() => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]stopDiscovery()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]stopDiscovery()=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _stopDiscovery_sub(adapter, timId));
	});
};
function _stopDiscovery_sub(adapter, timId){
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			if(! await adapter.isDiscovering()){
				utl.debugLog(DEBG_API, dbLv, DBG_NAME, "既にdiscovery停止中");
				utl.debugLog(DEBG_API, dbLv, DBG_NAME, "stopDiscovery完了");
			}else{
				utl.debugLog(DEBG_API, dbLv, DBG_NAME, "stopDiscovery開始");
				await adapter.stopDiscovery();
				utl.debugLog(DEBG_API, dbLv, DBG_NAME, "stopDiscovery完了");
			}

			clearTimeout(timId);
			ret.result = true;
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]stopDiscovery()=" + e);

			clearTimeout(timId);
			ret.result = false;
			ret.e = e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * node-bleの検出済デバイスリストの取得
 *  adapter      : bluetooth.defaultAdapter()
 *  timeout      : TimeOut時間
 ****************************************************************************/
NodeBleLib.prototype.getScanDevice = function(adapter, timeout){
	return new Promise(async (resolve) => {
		let timId = setTimeout(() => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]getScanDevice()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]getScanDevice()=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _getScanDevice_sub(adapter, timId));
	});
};
function _getScanDevice_sub(adapter, timId) {
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "devices開始");
			ret.devices = await adapter.devices();
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "devices完了");

			clearTimeout(timId);
			ret.result = true;
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]devices()=" + e);

			clearTimeout(timId);
			ret.result = false;
			ret.devices = [];
			ret.e = e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * node-bleのデバイス取得
 *  adapter      : bluetooth.defaultAdapter()
 *  bleId        : BLE Adress
 *  timeout      : TimeOut時間
 ****************************************************************************/
NodeBleLib.prototype.getDevice = function(adapter, bleId, timeout) {
	return new Promise(async (resolve) => {
		let timId = setTimeout(() => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]getDevice()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]getDevice()=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _getDevice_sub(adapter, bleId, timId));
	});
};
function _getDevice_sub(adapter, bleId, timId) {
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "getDevice開始");
			ret.device = await adapter.getDevice(bleId);
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "getDevice完了");

			clearTimeout(timId);
			ret.result = true;
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]getDevice()=" + e);

			clearTimeout(timId);
			ret.result = false;
			ret.device = null;
			ret.e = e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * localNameの取得
 *  device      : device object
 ****************************************************************************/
NodeBleLib.prototype.getName = function(device, retryLimit, errorInterval, timeout) {
	return new Promise(async (resolve) => {
		let ret = {};
		for(let i=0; i<retryLimit; i++){
			ret = {};
			ret = await _getName_to(device, timeout);
			if(ret.result === true){
				break;
			}else{
				if(retryLimit !== 1){
					await utl.waitFunc(errorInterval);
				}
			}
		}
		resolve(ret);
	});
};
function _getName_to(device, timeout) {
	return new Promise(async (resolve) => {
		let timId = setTimeout(async () => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]getName()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]getName()=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _getName_to_sub(device, timId));
	});
};
function _getName_to_sub(device, timId) {
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "getName開始");
			ret.name  = await device.getName();
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "getName完了:" + ret.name);

			clearTimeout(timId);
			ret.result = true;
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]getName()=" + e);

			clearTimeout(timId);
			ret.result = false;
			ret.name = null;
			ret.e = e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * aliaslNameの取得
 *  device      : device object
 ****************************************************************************/
NodeBleLib.prototype.getAlias = function(device, retryLimit, errorInterval, timeout) {
	return new Promise(async (resolve) => {
		let ret = {};
		for(let i=0; i<retryLimit; i++){
			ret = {};
			ret = await _getAlias_to(device, timeout);
			if(ret.result === true){
				break;
			}else{
				if(retryLimit !== 1){
					await utl.waitFunc(errorInterval);
				}
			}
		}
		resolve(ret);
	});
};
function _getAlias_to(device, timeout) {
	return new Promise(async (resolve) => {
		let timId = setTimeout(async () => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]getAlias()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]getAlias()=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _getAlias_to_sub(device, timId));
	});
};
function _getAlias_to_sub(device, timId) {
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "getAlias開始");
			ret.alias  = await device.getAlias();
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "getAlias完了:" + ret.alias);

			clearTimeout(timId);
			ret.result = true;
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]getAlias()=" + e);

			clearTimeout(timId);
			ret.result = false;
			ret.alias = null;
			ret.e = e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * RSSIの取得
 *  device      : device object
 ****************************************************************************/
NodeBleLib.prototype.getRssi = function(device, retryLimit, errorInterval, timeout) {
	return new Promise(async (resolve) => {
		let ret = {};
		for(let i=0; i<retryLimit; i++){
			ret = {};
			ret = await _getRssi_to(device, timeout);
			if(ret.result === true){
				break;
			}else{
				if(retryLimit !== 1){
					await utl.waitFunc(errorInterval);
				}
			}
		}
		resolve(ret);
	});
};
function _getRssi_to(device, timeout) {
	return new Promise(async (resolve) => {
		let timId = setTimeout(async () => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]getRssi()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]getRssi()=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _getRssi_to_sub(device, timId));
	});
};
function _getRssi_to_sub(device, timId) {
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "getRSSI開始");
			ret.rssi = await device.getRSSI();
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "getRSSI完了:" + ret.rssi);

			clearTimeout(timId);
			ret.result = true;
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]getRSSI()=" + e);

			clearTimeout(timId);
			ret.result = false;
			ret.rssi = null;
			ret.e = e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * removeDeviceの代替え機能
 *  bleId   : xx:xx:xx:xx:xx:xx形式
 *  timeout : TimeOut時間
 ****************************************************************************/
NodeBleLib.prototype.removeDevice = function(adapter, bleId, timeout) {
	return new Promise(async (resolve) => {
		let timId = setTimeout(() => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]removeDevice()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]removeDevice()=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _removeDevice_sub(adapter, bleId, timId));
	});
};
function _removeDevice_sub(adapter, bleId, timId) {
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			// bleIdはxx:xx:xx:xx:xx:xx形式
			let cmdExe = spawn('bluetoothctl', ['remove', bleId]);
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "removeDevice[" + bleId + "]開始");
			cmdExe.stdout.setEncoding('utf8');

			cmdExe.stderr.once('data', (err) => {
				let resString = err.toString();
				utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]removeDevice[" + bleId + "]=" + resString);
				clearTimeout(timId);

				ret.result = false;
				ret.e = resString;
				resolve(ret);
			});
			cmdExe.stdout.once('data', async (data) => {
				let resString = data.toString();			// Successful
				utl.debugLog(DEBG_API, dbLv, DBG_NAME, "removeDevice[" + bleId + "]完了=" + resString);
				clearTimeout(timId);

				ret.result = true;
				resolve(ret);
			});
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]removeDevice()=" + e);
			ret.result = false;
			clearTimeout(timId);

			ret.e = e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * manufacturerDataの取得
 *  device : device object
 *  timeout: TimeOut時間
 *  return : ex){ '76': Variant { signature: 'ay', value: [Buffer [Uint8Array]] } }
 ****************************************************************************/
NodeBleLib.prototype.getManufacturerData = function(device, timeout) {
	return new Promise(async (resolve) => {
		let timId = setTimeout(() => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]getManufacturerData()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]getManufacturerData()=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _getManufacturerData_sub(device, timId));
	});
};
function _getManufacturerData_sub(device, timId) {
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "manufacturerData開始");
			const rcvData = await device.manufacturerData();
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "manufacturerData完了:" + rcvData);
			clearTimeout(timId);
			ret.result = true;
			ret.manufacturerData = {};
			for(let keyKind in rcvData){
				let data = JSON.parse(JSON.stringify(rcvData[keyKind]));
				ret.manufacturerData[keyKind] = data.value.data;
			}
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]manufacturerData()=" + e);
			clearTimeout(timId);
			ret.result = false;
			ret.e = "[error]getManufacturerData()=" + e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * serviceDataの取得
 *  device : device object
 *  timeout: TimeOut時間
 *  return : ex){ '0000ffe1-0000-1000-8000-00805f9b34fb': Variant { signature: 'ay', value: [Buffer [Uint8Array]] } }
 ****************************************************************************/
NodeBleLib.prototype.getServiceData = function(device, timeout) {
	return new Promise(async (resolve) => {
		let timId = setTimeout(() => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]getServiceData()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]getServiceData()=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _getServiceData_sub(device, timId));
	});
};
function _getServiceData_sub(device, timId) {
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "serviceData開始");
			const rcvData = await device.serviceData();
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "serviceData完了:" + rcvData);
			clearTimeout(timId);
			ret.result = true;
			ret.serviceData = {};
			for(let keyKind in rcvData){
				let data = JSON.parse(JSON.stringify(rcvData[keyKind]));
				ret.serviceData[keyKind] = data.value.data;
			}
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]serviceData()=" + e);
			clearTimeout(timId);
			ret.result = false;
			ret.e = "[error]getServiceData()=" + e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * node-bleの接続
 *  device        : device object
 *  retryLimit    : リトライ回数
 *  errorInterval : リトライ間隔
 *  timeout       : TimeOut時間
 ****************************************************************************/
NodeBleLib.prototype.connect = function(device, retryLimit, errorInterval, timeout) {
	return new Promise(async (resolve) => {
		let ret = {};
		for(let i=0; i<retryLimit; i++){
			ret = {};
			ret = await _connect_to(device, timeout);
			if(ret.result){
				break;
			}else{
				if(retryLimit !== 1){
					await utl.waitFunc(errorInterval);
				}
			}
		}
		resolve(ret);
	});
};
function _connect_to(device, timeout) {
	return new Promise(async (resolve) => {
		let timId = setTimeout(async () => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]connect()=timeout");
			await device.disconnect();
			let ret = {};
			ret.result = false;
			ret.e = "[error]connect()=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _connect_to_sub(device, timId));
	});
};
function _connect_to_sub(device, timId) {
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			if(! await device.isConnected()){
				utl.debugLog(DEBG_API, dbLv, DBG_NAME, "デバイス接続開始");
				await device.connect();
				utl.debugLog(DEBG_API, dbLv, DBG_NAME, "デバイス接続完了");
				ret.result = true;
			}else{
				utl.debugLog(DEBG_API, dbLv, DBG_NAME, "デバイス接続済");
				await device.disconnect();
				utl.debugLog(DEBG_API, dbLv, DBG_NAME, "デバイス接続NG(接続済なので一旦disconnect");
				ret.result = false;
			}

			clearTimeout(timId);
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]connect()=" + e);

			clearTimeout(timId);
			await device.disconnect();
			ret.result = false;
			ret.e = e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * node-bleの切断
 *  device        : device object
 *  retryLimit    : リトライ回数
 *  errorInterval : リトライ間隔
 *  timeout       : TimeOut時間
 ****************************************************************************/
NodeBleLib.prototype.disconnect = function(device, retryLimit, errorInterval, timeout) {
	return new Promise(async (resolve) => {
		let ret = {};
		for(let i=0; i<retryLimit; i++){
			ret = {};
			ret = await _disconnect_to(device, timeout);
			if(ret.result === true){
				break;
			}else{
				if(retryLimit !== 1){
					await utl.waitFunc(errorInterval);
				}
			}
		}
		resolve(ret);
	});
};
function _disconnect_to(device, timeout) {
	return new Promise(async (resolve) => {
		let timId = setTimeout(async () => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]disconnect()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]disconnect()=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _disconnect_sub(device, timId));
	});
};
function _disconnect_sub(device, timId){
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			if(await device.isConnected()){
				utl.debugLog(DEBG_API, dbLv, DBG_NAME, "デバイス切断開始");
				await device.disconnect();
				utl.debugLog(DEBG_API, dbLv, DBG_NAME, "デバイス切断完了");
			}else{
				utl.debugLog(DEBG_API, dbLv, DBG_NAME, "デバイス切断済");
			}

			clearTimeout(timId);
			ret.result = true;
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]disconnect()=" + e);

			clearTimeout(timId);
			ret.result = false;
			ret.e = e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * node-bleのGattServer
 *  device        : device object
 *  retryLimit    : リトライ回数
 *  errorInterval : リトライ間隔
 *  timeout       : TimeOut時間
 ****************************************************************************/
NodeBleLib.prototype.startGattServer = function(device, retryLimit, errorInterval, timeout) {
	return new Promise(async (resolve) => {
		let ret = {};
		for(let i=0; i<retryLimit; i++){
			ret = {};
			ret = await _startGattServer_to(device, timeout);
			if(ret.result === true){
				break;
			}else{
				if(retryLimit !== 1){
					await utl.waitFunc(errorInterval);
				}
			}
		}
		resolve(ret);
	});
};
function _startGattServer_to(device, timeout) {
	return new Promise(async (resolve) => {
		let timId = setTimeout(async () => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]gatt()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]gatt()=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _startGattServer_sub(device, timId));
	});
};
function _startGattServer_sub(device, timId){
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "GattServer開始");
			ret.gattServer = await device.gatt();
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "GattServer開始完了");

			clearTimeout(timId);
			ret.result = true;
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]startGattServer()=" + e);

			clearTimeout(timId);
			ret.result = false;
			ret.gattServer = null;
			ret.e = e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * node-bleのPrimaryService
 *  gattServer    : gattServer object
 *  serviceUuid   : SERVICE_UUID
 *  retryLimit    : リトライ回数
 *  errorInterval : リトライ間隔
 *  timeout       : TimeOut時間
 ****************************************************************************/
NodeBleLib.prototype.getPrimaryService = function(gattServer, serviceUuid, retryLimit, errorInterval, timeout) {
	return new Promise(async (resolve) => {
		let ret = {};
		for(let i=0; i<retryLimit; i++){
			ret = {};
			ret = await _getPrimaryService_to(gattServer, serviceUuid, timeout);
			if(ret.result === true){
				break;
			}else{
				if(retryLimit !== 1){
					await utl.waitFunc(errorInterval);
				}
			}
		}
		resolve(ret);
	});
};
function _getPrimaryService_to(gattServer, serviceUuid, timeout){
	return new Promise(async (resolve) => {
		let timId = setTimeout(async () => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]getPrimaryService()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]getPrimaryService()=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _getPrimaryService_sub(gattServer, serviceUuid, timId));
	});
};
function _getPrimaryService_sub(gattServer, serviceUuid, timId){
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "primaryService取得開始");
			ret.primaryService = await gattServer.getPrimaryService(serviceUuid);
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "primaryService取得完了");

			clearTimeout(timId);
			ret.result = true;
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]getPrimaryService()=" + e);

			clearTimeout(timId);
			ret.result = false;
			ret.primaryService = null;
			ret.e = e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * node-bleのCharacteristic取得
 *  primaryService     : primaryService object
 *  characteristicUuid : CHARACTERISTIC_UUID
 *  retryLimit         : リトライ回数
 *  errorInterval      : リトライ間隔
 *  timeout            : TimeOut時間
 ****************************************************************************/
NodeBleLib.prototype.getCharacteristic = function(primaryService, characteristicUuid, retryLimit, errorInterval, timeout) {
	return new Promise(async (resolve) => {
		let ret = {};
		for(let i=0; i<retryLimit; i++){
			ret = {};
			ret = await _getCharacteristic_to(primaryService, characteristicUuid, timeout);
			if(ret.result === true){
				break;
			}else{
				if(retryLimit !== 1){
					await utl.waitFunc(errorInterval);
				}
			}
		}
		resolve(ret);
	});
};
function _getCharacteristic_to(primaryService, characteristicUuid, timeout) {
	return new Promise(async (resolve) => {
		let timId = setTimeout(async () => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]getCharacteristic()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]getCharacteristic()=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _getCharacteristic_sub(primaryService, characteristicUuid, timId));
	});
};
function _getCharacteristic_sub(primaryService, characteristicUuid, timId){
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "characteristic取得開始");
			ret.characteristic = await primaryService.getCharacteristic(characteristicUuid);
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "characteristic取得完了");

			clearTimeout(timId);
			ret.result = true;
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]getCharacteristic()=" + e);

			clearTimeout(timId);
			ret.result = false;
			ret.characteristic = null;
			ret.e = e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * characteristic指定でのwrite
 * mode
 *  "command"  : Write without response
 *  "request"  : Write with response
 *  "reliable" : Reliable Write
 *  "timeout"  : API Timeout value.
 ****************************************************************************/
NodeBleLib.prototype.writeCharacteristic = function(characteristic, buffer, mode, timeout) {
	return new Promise(async (resolve) => {
		let timId = setTimeout(() => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]writeValue()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]writeValue()=timeout";
			resolve(ret);
		}, timeout);
		utl.debugLog(DEBG_API, dbLv, DBG_NAME, "writeCharacteristic() [" + buffer.toString('hex') + "]");
		resolve(await _writeCharacteristic_sub(characteristic, buffer, mode, timId));
	});
};
function _writeCharacteristic_sub(characteristic, buffer, mode, timId){
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			let op = {};
			op.type = mode;
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "writeValue開始[" + mode + "], buffer=" + buffer.toString('hex'));
			await characteristic.writeValue(buffer, op);
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "writeValue完了");

			clearTimeout(timId);
			ret.result = true;
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]writeValue()=" + e);

			clearTimeout(timId);
			ret.result = false;
			ret.e = e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * characteristic指定でのread
 *  "characteristic" : characteristic object
 *  "timeout"        : API Timeout value.
 ****************************************************************************/
NodeBleLib.prototype.readCharacteristic = function(characteristic, timeout) {
	return new Promise(async (resolve) => {
		let timId = setTimeout(() => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]readValue()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]readValue()=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _readCharacteristic_sub(characteristic, timId));
	});
};
function _readCharacteristic_sub(characteristic, timId){
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "readValue開始");
			let retPayload = await characteristic.readValue(0);
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "readValue完了");

			clearTimeout(timId);
			ret.result = true;
			ret.payload = retPayload;
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]readValue()=" + e);

			clearTimeout(timId);
			ret.result = false;
			ret.payload = null;
			ret.e = e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * Notificationsのstart
 *  "characteristic" : characteristic object
 *  retryLimit       : リトライ回数
 *  errorInterval    : リトライ間隔
 *  timeout          : TimeOut時間
 ****************************************************************************/
NodeBleLib.prototype.startNotifications = function(characteristic, retryLimit, errorInterval, timeout) {
	return new Promise(async (resolve) => {
		let ret = {};
		for(let i=0; i<retryLimit; i++){
			ret = {};
			ret = await _startNotifications_to(characteristic, timeout);
			if(ret.result === true){
				break;
			}else{
				if(retryLimit !== 1){
					await utl.waitFunc(errorInterval);
				}
			}
		}
		resolve(ret);
	});
};
function _startNotifications_to(characteristic, timeout) {
	return new Promise(async (resolve) => {
		let timId = setTimeout(async () => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]startNotifications()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]startNotifications()=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _startNotifications_sub(characteristic, timId));
	});
};
function _startNotifications_sub(characteristic, timId){
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "startNotifications開始");
			await characteristic.startNotifications();
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "startNotifications完了");

			clearTimeout(timId);
			ret.result = true;
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]startNotifications()=" + e);

			clearTimeout(timId);
			ret.result = false;
			ret.e = e;
			resolve(ret);
		}
	});
};

/*****************************************************************************
 * Notificationsのstop
 *  "characteristic" : characteristic object
 *  timeout          : TimeOut時間
 ****************************************************************************/
NodeBleLib.prototype.stopNotifications = function(characteristic, timeout) {
	return new Promise(async (resolve) => {
		let timId = setTimeout(() => {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]stopNotifications()=timeout");
			let ret = {};
			ret.result = false;
			ret.e = "[error]stopNotifications()=timeout";
			resolve(ret);
		}, timeout);
		resolve(await _stopNotifications_sub(characteristic, timId));
	});
};
function _stopNotifications_sub(characteristic, timId){
	return new Promise(async (resolve) => {
		let ret = {};
		try {
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "stopNotifications開始");
			await characteristic.stopNotifications();
			utl.debugLog(DEBG_API, dbLv, DBG_NAME, "stopNotifications完了");

			clearTimeout(timId);
			ret.result = true;
			resolve(ret);
		} catch (e) {
			utl.debugLog(DEBG_ERR, dbLv, DBG_NAME, "[error]stopNotifications()=" + e);

			clearTimeout(timId);
			ret.result = false;
			ret.e = e;
			resolve(ret);
		}
	});
};

module.exports = NodeBleLib;
