/* ------------------------------------------------------------------
 * node-gotapi - node-gotapi-plugin-environment-btpm25v/index.js
 *                for RATOC Systems
 * Copyright (c) 2022, NTT DoCoMo, All rights reserved.
 * Date: 2022-07-19
 * node-module : fs, path, node-ble
 * ---------------------------------------------------------------- */
'use strict';
// node lib
const mPath = require('path');
const mFs = require('fs');


// Define
const DEBG_LV1 = 1;
const DEBG_LV2 = 2;
const DEBG_LV3 = 3;
const WAIT_TIME_OUT       = 7500;
const WAIT_TIME_OUT_10MS  = 10;
const WAIT_TIME_OUT_100MS = 100;
const WAIT_TIME_OUT_500MS = 500;
const WAIT_TIME_OUT_1SEC  = 1000;
const WAIT_TIME_OUT_3SEC  = 3000;
const WAIT_TIME_OUT_10SEC = 10000;
const WAIT_TIME_OUT_20SEC = 20000;
const WAIT_TIME_OUT_30SEC = 30000;
const WAIT_TIME_OUT_1MIN  = 60000;
const WAIT_TIME_OUT_MEASU = 40000;
const WAIT_TIME_OUT_API   = 20000;
const READ_LAST_DATA      = "F000000000000000000000000000000000000000";
const ERROR_REBOOT_LIMIT  = 10;

// config情報
const CONFIG = JSON.parse(mFs.readFileSync(mPath.join(__dirname, 'config.json'), 'utf8'));
const SERVICE_UUID                           = CONFIG.SERVICE_UUID;
const CHARACTERISTIC_AUTO_MEASUREMENT_NOTIFY = CONFIG.CHARACTERISTIC_AUTO_MEASUREMENT_NOTIFY;
const CHARACTERISTIC_AUTO_MEASUREMENT        = CONFIG.CHARACTERISTIC_AUTO_MEASUREMENT;
const CHARACTERISTIC_SET_DATETIME            = CONFIG.CHARACTERISTIC_SET_DATETIME;
const CHARACTERISTIC_ONESHOT_MEASUREMENT     = CONFIG.CHARACTERISTIC_ONESHOT_MEASUREMENT;
const CHARACTERISTIC_LATEST_MEASUREMENT      = CONFIG.CHARACTERISTIC_LATEST_MEASUREMENT;
const DEVICE_KIND                            = CONFIG.DEVICE_KIND;
const LOCALNAME                              = CONFIG.LOCALNAME;
const DEBUG_LEVEL                            = CONFIG.DEBUG_LEVEL;
const SERVICE_ID_BASS_INF                    = CONFIG.SERVICE_ID_BASS_INF;
const START_DELAY_TIME                       = CONFIG.START_DELAY_TIME * WAIT_TIME_OUT_1SEC;
const ERROR_INTERVAL                         = CONFIG.ERROR_INTERVAL * WAIT_TIME_OUT_1SEC;
const RETRY_LIMIT                            = CONFIG.RETRY_LIMIT;
const SCAN_INTERVAL                          = CONFIG.SCAN_INTERVAL * WAIT_TIME_OUT_1SEC;
const BLE_LIB_TIMEOUT                        = CONFIG.BLE_LIB_TIMEOUT * WAIT_TIME_OUT_1SEC;

// local lib
let utl = new (require('./utilsFunc.js'))();
let nodeBleLib = new (require('./bleLib_node-ble.js'))(utl, DEBUG_LEVEL, DEVICE_KIND, "node-ble");

// RAM
let knownDeviceList = {};
let deviceList = [];
let UNIQ_PLG_ID = utl.uniqueIdGene();

// ble処理用RAM
let _eventTimerTable = {};
let _errorCounr = 0;

/*********************************************************
* serviceId生成
* 
*********************************************************/
function serviceIdGenerator(bassString, bleId) {
	let sensorId = "";
	sensorId = bleId;
	sensorId = sensorId.replace(/ /g, '');
	sensorId = sensorId.replace(/:/g, '');
	sensorId = sensorId.replace(/-/g, '');
	sensorId = sensorId.toLowerCase();

	let retString  = "";
	retString += bassString;
	retString += "-";
	retString += sensorId;
	return retString;
};

/*********************************************************
* Init
* 全部登録
*********************************************************/
function deviceInitReg() {
	return new Promise(async function(resolve){
		const filePath = __dirname + '/devList.json'
		let fileChk = utl.existenceCheck(filePath);
		if(fileChk === true){
			deviceList = JSON.parse(mFs.readFileSync(filePath, 'utf8'));
		}else{
			// ファイルが無いので空で作成する
			fileChk = await utl.writeFile(filePath, JSON.stringify(deviceList));
		}

		// デバイスリストの情報で初期リスト作成
		for(let i=0; i<deviceList.length; i++){
			deviceReg(deviceList[i]);
		}
		resolve(true);
	});
};

/*********************************************************
* Init
* 個別登録
*********************************************************/
function deviceReg(deviceListInfo) {
	return new Promise(async (resolve) => {
		let sensorId = "";
		sensorId = deviceListInfo.bleId;
		sensorId = sensorId.replace(/ /g, '');
		sensorId = sensorId.replace(/:/g, '');
		sensorId = sensorId.replace(/-/g, '');
		sensorId = sensorId.toLowerCase();

		let serviceId = serviceIdGenerator(SERVICE_ID_BASS_INF, sensorId);
		let timeStamp = (new Date).getTime();
		let timeStampString = utl.calNowDateTime(timeStamp);

		let tmpDevInf = {};
		tmpDevInf.serviceId = serviceId;
		tmpDevInf.systemId  = sensorId;
		tmpDevInf.connected = false;
		tmpDevInf.timeStamp = timeStamp;
		tmpDevInf.timeStampString = timeStampString;
		tmpDevInf.timingStamp = 0;
		tmpDevInf.measurCycSet = false;
		tmpDevInf.measurInterval = deviceListInfo.measurInterval;

		tmpDevInf.environment = {};

		tmpDevInf.atmosphericPressure = {};
		tmpDevInf.atmosphericPressure.value = -1;

		tmpDevInf.humidity = {};
		tmpDevInf.humidity.value = -1;

		tmpDevInf.illuminance = {};
		tmpDevInf.illuminance.value = -1;

		tmpDevInf.pm10 = {};
		tmpDevInf.pm10.value = -1;

		tmpDevInf.pm25 = {};
		tmpDevInf.pm25.value = -1;

		tmpDevInf.temperature = {};
		tmpDevInf.temperature.value = -999;

		tmpDevInf.voc = {};
		tmpDevInf.voc.value = -1;

		tmpDevInf.uv = {};
		tmpDevInf.uv.value = -1;

		tmpDevInf.co2Concentration = {};
		tmpDevInf.co2Concentration.value = -1;

		// RAM更新
		knownDeviceList[serviceId] = {};
		knownDeviceList[serviceId] = tmpDevInf;

		if(retIniReg[serviceId] === undefined){
			retIniReg[serviceId] = null;
		}
		retIniReg[serviceId] = await nodeBleLib.startLib(BLE_LIB_TIMEOUT);
		if(retIniReg[serviceId].result){
			// OK
			let retDi = await nodeBleLib.startDiscovery(retIniReg[serviceId].adapter, BLE_LIB_TIMEOUT);
			if(retDi.result){
				// OK
				resolve(true);
			}else{
				// NG
				retIniReg[serviceId] = null;
				resolve(false);
			}
		}else{
			// NG
			retIniReg[serviceId] = null;
			resolve(false);
		}
	});
};



/********************************************
 * Device依存処理群
 *  
 *  
 ********************************************/
/*********************************************************
* DeviceTimer function
*********************************************************/
function _startEventTimer(bleId, delay, callFunc) {
	let delayVal = 0;
	if (delay === undefined) {
		delayVal = WAIT_TIME_OUT;
	}else{
		delayVal = delay;
	}
	_clearEventTimer(bleId);
	_eventTimerTable[bleId] = setTimeout(async (bleIdTmp, callFuncTmp) => {
									if (callFuncTmp && (typeof callFuncTmp === 'function')){
										// 指定の処理があれば実行する
										utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "Time Out! [" + bleIdTmp + "]");

										await callFuncTmp();
									}
								}, delayVal, bleId, callFunc);
};
function _clearEventTimer(bleId) {
	let timerId = _eventTimerTable[bleId];
	if (timerId !== undefined) {
		clearTimeout(timerId);
		_eventTimerTable[bleId] = -1;
	}
};

/************************************************
 〇'node-ble'
 
 ***********************************************/
let retIniReg = {};
function _startDeviceFinder() {
	return new Promise(async (resolve) => {
		for(let serviceIdKey in knownDeviceList){
			if(retIniReg[serviceIdKey] === null){
				retIniReg[serviceIdKey] = await nodeBleLib.startLib(BLE_LIB_TIMEOUT);
				if(retIniReg[serviceIdKey].result){
					let retDis = await nodeBleLib.startDiscovery(retIniReg[serviceIdKey].adapter, BLE_LIB_TIMEOUT);
					if(retDis.result){
						await utl.waitFunc(SCAN_INTERVAL);
					}else{
						// NG
						utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "処理完了なのでdestroyします 0:      [" + serviceIdKey + "] ");
						await nodeBleLib.destroy(retIniReg[serviceIdKey].adapter, retIniReg[serviceIdKey].destroy, BLE_LIB_TIMEOUT);
						retIniReg[serviceIdKey] = null;
						await utl.waitFunc(ERROR_INTERVAL);
					}
				}else{
					// NG
					utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "startLib(BLE_LIB_TIMEOUT) ERROR");
					retIniReg[serviceIdKey] = null;
					await utl.waitFunc(ERROR_INTERVAL);
				}
			}else{
				let retGetReg = await nodeBleLib.getScanDevice(retIniReg[serviceIdKey].adapter, BLE_LIB_TIMEOUT + SCAN_INTERVAL);
				if(retGetReg.result){

utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "1-----------------------------------------------");
utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, JSON.stringify(retGetReg.devices));
utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "1-----------------------------------------------");

					for(let i=0; i<retGetReg.devices.length; i++){
						// Scan結果のBLE IDが知っているIDを含んでいるかを確認。
						let hogeFlg = false;
						const bleId = retGetReg.devices[i];
						const serviceId = serviceIdGenerator(SERVICE_ID_BASS_INF, bleId);
						if(serviceIdKey === serviceId){
							// 管理対象のデバイスだった
							if(knownDeviceList[serviceIdKey].connected === false){
								// 測定タイミング調整
								let goFlg = false;
								let remainingTime = 0;
								if(knownDeviceList[serviceIdKey].timingStamp === 0){
									// 即時実行
									goFlg = true;
								}else{
									remainingTime = knownDeviceList[serviceIdKey].measurInterval - ((new Date).getTime() - knownDeviceList[serviceIdKey].timingStamp) + WAIT_TIME_OUT_MEASU;
									if(remainingTime <= 0){
										goFlg = true;
									}else{
										goFlg = false;
									}
								}

								if(goFlg){
									// 測定タイミングなので測定を開始する
									utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "[" + bleId + "]Timing:Just now!");

									let retDevice = await nodeBleLib.getDevice(retIniReg[serviceIdKey].adapter, bleId, BLE_LIB_TIMEOUT);
									if(retDevice.result){
										// 接続(同期処理) -> 読み出し(同期処理)
										let retCnt = await _connectDevice(retIniReg[serviceIdKey].adapter, retDevice.device, serviceIdKey, bleId, RETRY_LIMIT, ERROR_INTERVAL);
										if(retCnt){
											let retRad = await _readDevice(retIniReg[serviceIdKey].adapter, retDevice.device, serviceIdKey, bleId, RETRY_LIMIT, ERROR_INTERVAL);
											if(retRad){
												// OK
											}else{
												utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "Device アクセスエラー[_readDevice()]");
											}
										}else{
											utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "Device 接続エラー[_connectDevice()]");
										}

										if(_errorCounr > ERROR_REBOOT_LIMIT){
											console.error('Plugin[node-gotapi-plugin-environment-btpm25v-v2]BLE接続の連続エラー回数が閾値を超えました。プロセス再起動を行います。');
											utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "Plugin[node-gotapi-plugin-environment-btpm25v-v2]BLE接続の連続エラー回数が閾値を超えました。プロセス再起動を行います。");
											process.exit();
										}else{
											utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "処理完了 :      [" + bleId + "]");
										}
									}else{
										_errorCounr++;
										utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "Device取得エラー:" + retDevice.e);
									}
								}else{
									utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "登録デバイスだが今回はタイミングではないので無視する。[" + bleId + ", 残り = " + remainingTime/1000 + "秒]");
								}
							}else{
								utl.debugLog(DEBG_LV3, DEBUG_LEVEL, DEVICE_KIND, "処理中なのでSkipします :      [" + bleId + "]");
							}
						}else{
							utl.debugLog(4, DEBUG_LEVEL, DEVICE_KIND, "対象外デバイスなので無視する。[" + bleId + "]");
						}
					}
					if(retGetReg.devices.length === 0){
						let retPw = await nodeBleLib.isPowered(retIniReg[serviceIdKey].adapter, BLE_LIB_TIMEOUT);
						if(retPw.result){
							if(retPw.isPowered){
								utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "タイミングだが検出できないのでSkipする。[" + serviceIdKey + "]");

utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "2===============================================");
utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, JSON.stringify(retGetReg.devices));
utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "2===============================================");

							}else{
								utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "PowerOff検出。PowerOnに変更します。[" + serviceIdKey + "]");
								await nodeBleLib.adapterPower(true, BLE_LIB_TIMEOUT);
							}
						}else{
							utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "PowerOff検出失敗。[" + serviceIdKey + "]" + retPw.e);
						}
					}
				}else{
					// Lib error
					utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "SCANエラー:" + retGetReg.e);
				}

				// 廃棄可能
				utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "処理完了なのでdestroyします:      [" + serviceIdKey + "] ");
				await nodeBleLib.destroy(retIniReg[serviceIdKey].adapter, retIniReg[serviceIdKey].destroy, BLE_LIB_TIMEOUT);
				retIniReg[serviceIdKey] = null;
			}
		}

		setTimeout(async () => {
			await _startDeviceFinder();
		}, WAIT_TIME_OUT_500MS);

		resolve(true);
	});
};

/************************************************
 * _disconnect
 *  
 *  
 ***********************************************/
function _disconnect(device, serviceId, bleId, retryLimit, errorInterval) {
	return new Promise(async (resolve) => {
		if(await device.isConnected()){
			utl.debugLog(DEBG_LV2, DEBUG_LEVEL, DEVICE_KIND, "デバイス切断開始[" + bleId + "]");

			let retDis = await nodeBleLib.disconnect(device, retryLimit, errorInterval, BLE_LIB_TIMEOUT);
			if(retDis.result){
				utl.debugLog(DEBG_LV2, DEBUG_LEVEL, DEVICE_KIND, "デバイス切断完了[" + bleId + "]");
				resolve(true);
			}else{
				// 切断NG
				utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "[error]nodeBleLib.disconnect()=" + retDis.e);
				resolve(false);
			}
		}else{
			utl.debugLog(DEBG_LV2, DEBUG_LEVEL, DEVICE_KIND, "既にデバイス切断済[" + bleId + "]");
			resolve(true);
		}
	});
};

/************************************************
 * _connectDevice
 *  
 ***********************************************/
function _connectDevice(adapterTmp, deviceTmp, serviceIdTmp, bleIdTmp, retryLimitTmp, errorIntervalTmp){
	return new Promise(async (resolve) => {
		let adapter = adapterTmp;
		let device = deviceTmp;
		let serviceId = serviceIdTmp;
		let bleId = bleIdTmp;
		let retryLimit = retryLimitTmp;
		let errorInterval = errorIntervalTmp;

		// 不明なデバイスへコネクトした場合は処理無し(登録デバイスのみ対応の為、ここには来ないはず)
		if(knownDeviceList[serviceId] === undefined){
			resolve(false);
		}else{
			knownDeviceList[serviceId].connected = true;

			utl.debugLog(DEBG_LV3, DEBUG_LEVEL, DEVICE_KIND, "[start]nodeBleLib.connect()=" + bleId);
			let retCnt = await nodeBleLib.connect(device, retryLimit, errorInterval, BLE_LIB_TIMEOUT);
			if(retCnt.result){
				// 接続OK
				utl.debugLog(DEBG_LV2, DEBUG_LEVEL, DEVICE_KIND, "[ok]nodeBleLib.connect()=" + bleId);
				resolve(true);
			}else{
				// 接続NG
				utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "[error]nodeBleLib.connect()=" + retCnt.e);

				// Errorカウンタ
				_errorCounr++;

				// 切断実施(接続できていないならエラーが返ってくるので気にせず実施)
				await _disconnect(device, serviceId, bleId, retryLimit, errorInterval);

				// REXが再起動していたりして測定周期がクリアされている可能性があるので
				// 次回は再度測定周期の再設定から実施させる
				knownDeviceList[serviceId].measurCycSet = false;

				// 処理の終わった物をSystemから忘れさせる(自分の管理対象以外には実施しない事))
				await nodeBleLib.removeDevice(adapter, bleId, BLE_LIB_TIMEOUT);

				await utl.waitFunc(errorInterval);

				knownDeviceList[serviceId].connected = false;
				resolve(false);
			}
		}
	});
};

/************************************************
 *  _readDevice
 *    ->getName()
 *    ->_discoverServices()
 *    ->_disconnect()
 *    ->removeDevice()
 *    ->connected = false
 *  
 ***********************************************/
function _readDevice(adapterTmp, deviceTmp, serviceIdTmp, bleIdTmp, retryLimitTmp, errorIntervalTmp){
	return new Promise(async (resolve) => {
		let adapter = adapterTmp;
		let device = deviceTmp;
		let serviceId = serviceIdTmp;
		let bleId = bleIdTmp;
		let retryLimit = retryLimitTmp;
		let errorInterval = errorIntervalTmp;

		if(knownDeviceList[serviceId] === undefined){
			// 不明なデバイスへコネクトした場合は処理無し(登録デバイスのみ対応の為、ここには来ないはず)
			utl.debugLog(DEBG_LV2, DEBUG_LEVEL, DEVICE_KIND, bleId + ":は未登録、処理無");

			// 切断実施
			await _disconnect(device, serviceId, bleId, retryLimit, errorInterval);

			// 処理の終わった物をSystemから忘れさせる(自分の管理対象以外には実施しない事))
			await nodeBleLib.removeDevice(adapter, bleId, BLE_LIB_TIMEOUT);

			knownDeviceList[serviceId].connected = false;
			resolve(false);
		}else{
			// 処理が必要
			let ret = await _discoverServices(adapter, device, serviceId, bleId, retryLimit, errorInterval);

			// 切断実施
			await _disconnect(device, serviceId, bleId, retryLimit, errorInterval);

			// 処理の終わった物をSystemから忘れさせる(自分の管理対象以外には実施しない事))
			await nodeBleLib.removeDevice(adapter, bleId, BLE_LIB_TIMEOUT);

			// Errorカウンタ
			if(ret){
				_errorCounr = 0;
			}else{
				_errorCounr++;
				await utl.waitFunc(errorInterval);
			}

			knownDeviceList[serviceId].connected = false;
			resolve(ret);
		}
	});
};

function _discoverServices(adapter, device, serviceId, bleId, retryLimit, errorInterval){
	return new Promise(async (resolve) => {
		let flgLimit = false;

		// ------------ GattSrv生成 ---------------------------------------------------------------------------------------------------
		let retGattSrv = await nodeBleLib.startGattServer(device, retryLimit, errorInterval, BLE_LIB_TIMEOUT);
		if(retGattSrv.result === false){
			utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "[error]nodeBleLib.startGattServer()=" + bleId);
			resolve(false);
			return;
		}
		let retService = await nodeBleLib.getPrimaryService(retGattSrv.gattServer, SERVICE_UUID, retryLimit, errorInterval, BLE_LIB_TIMEOUT);
		if(retService.result === false){
			utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "[error]nodeBleLib.getPrimaryService()=" + SERVICE_UUID);
			resolve(false);
			return;
		}

		// ------------ データアクセス ---------------------------------------------------------------------------------------------------
		if(knownDeviceList[serviceId].measurCycSet === false){
			// -------- 現在時間を設定 ---------------------------------------------------------------------------------------------------
			let dateTimeRequest = await nodeBleLib.getCharacteristic(retService.primaryService, CHARACTERISTIC_SET_DATETIME, retryLimit, errorInterval, BLE_LIB_TIMEOUT);
			if(dateTimeRequest.result === false){
				utl.debugLog(1, DEBUG_LEVEL, DEVICE_KIND, "[error]nodeBleLib.getCharacteristic()=" + CHARACTERISTIC_SET_DATETIME);
				knownDeviceList[serviceId].measurCycSet = false;
				knownDeviceList[serviceId].timingStamp = 0;
				resolve(false);
				return;
			}
			let retWriteTim;
			const buf0 = Buffer.from('01' + getHexDatetime(new Date()), 'hex');
			flgLimit = true;
			utl.debugLog(1, DEBUG_LEVEL, DEVICE_KIND, "現在時間設定");
			for(let i=0; i<retryLimit; i++){
				retWriteTim = await nodeBleLib.writeCharacteristic(dateTimeRequest.characteristic, buf0, "request", WAIT_TIME_OUT_API);
				if(retWriteTim.result === true){
					flgLimit = false;
					break;
				}else{
					await utl.waitFunc(errorInterval);
				}
			}
			if(flgLimit){
				utl.debugLog(1, DEBUG_LEVEL, DEVICE_KIND, "[error]nodeBleLib.writeCharacteristic(" + CHARACTERISTIC_SET_DATETIME + ")=" + buf0.toString());
				knownDeviceList[serviceId].measurCycSet = false;
				knownDeviceList[serviceId].timingStamp = 0;
				resolve(false);
				return;
			}
			// -------- 現在時間を設定 ---------------------------------------------------------------------------------------------------

			// -------- Autoデータ収集周期を設定 -----------------------------------------------------------------------------------------
			let autoMesurRequest = await nodeBleLib.getCharacteristic(retService.primaryService, CHARACTERISTIC_AUTO_MEASUREMENT, retryLimit, errorInterval, BLE_LIB_TIMEOUT);
			if(autoMesurRequest.result === false){
				utl.debugLog(1, DEBUG_LEVEL, DEVICE_KIND, "[error]nodeBleLib.getCharacteristic()=" + CHARACTERISTIC_AUTO_MEASUREMENT);
				knownDeviceList[serviceId].measurCycSet = false;
				knownDeviceList[serviceId].timingStamp = 0;
				resolve(false);
				return;
			}
			const dt = new Date();
			dt.setMinutes(dt.getMinutes() + 1);
			// インターバルを16進数に変換
			let intervalTim = (knownDeviceList[serviceId].measurInterval / WAIT_TIME_OUT_1MIN);
			let it = ('000' + intervalTim.toString(16)).slice(-4);
			it = it.slice(2) + it.slice(0, 2);
			const buf1 = Buffer.from('01' + it + getHexDatetime(dt), 'hex');
			let retWriteMesur;
			flgLimit = true;
			utl.debugLog(1, DEBUG_LEVEL, DEVICE_KIND, "Autoデータ収集周期設定");
			for(let i=0; i<retryLimit; i++){
				retWriteMesur = await nodeBleLib.writeCharacteristic(autoMesurRequest.characteristic, buf1, "request", WAIT_TIME_OUT_API);
				if(retWriteMesur.result === true){
					flgLimit = false;
					break;
				}else{
					await utl.waitFunc(errorInterval);
				}
			}
			if(flgLimit){
				utl.debugLog(1, DEBUG_LEVEL, DEVICE_KIND, "[error]nodeBleLib.writeCharacteristic(" + CHARACTERISTIC_AUTO_MEASUREMENT + ")=" + buf1.toString());
				knownDeviceList[serviceId].measurCycSet = false;
				knownDeviceList[serviceId].timingStamp = 0;
				resolve(false);
				return;
			}
			// -------- Autoデータ収集周期を設定 -----------------------------------------------------------------------------------------

			// 最初の計測開始時間までの調整値を開始時間に設定(初回は1分後に取得するようにする)
			knownDeviceList[serviceId].timingStamp = (new Date).getTime() - (knownDeviceList[serviceId].measurInterval - WAIT_TIME_OUT_1MIN);

			// デバイスからデータ読み出しの初期設定が済
			knownDeviceList[serviceId].measurCycSet = true;
		}else{
			// -------- Lastデータ収集 ---------------------------------------------------------------------------------------------------
			let lastMesurRequest = await nodeBleLib.getCharacteristic(retService.primaryService, CHARACTERISTIC_LATEST_MEASUREMENT, retryLimit, errorInterval, BLE_LIB_TIMEOUT);
			if(lastMesurRequest.result === false){
				utl.debugLog(1, DEBUG_LEVEL, DEVICE_KIND, "[error]nodeBleLib.getCharacteristic()=" + CHARACTERISTIC_LATEST_MEASUREMENT);
				// 次のタイミングに任せる
				resolve(true);
				return;
			}
			await _getAutoMeasuredData(lastMesurRequest.characteristic, serviceId, bleId, retryLimit, errorInterval);
			// -------- Lastデータ収集 ---------------------------------------------------------------------------------------------------
		}
		resolve(true);
	});
};

function _getAutoMeasuredData(characteristic, serviceId, bleId, retryLimit, errorInterval){
	return new Promise(async (resolve) => {
		let flgLimit = false;

		try{
			utl.debugLog(1, DEBUG_LEVEL, DEVICE_KIND, "notify 受信設定");
			characteristic.once('valuechanged', async(buffer) => {
				// 取得したデータで上書き
				const retDat = convertMeasurementData(buffer, 20);

				let timeStampString = getDateTime(retDat);
				if(timeStampString === "2000-00-00T00:00:00+09:00"){
					// 無効なデータ:デバイスからデータ読み出しの初期設定が必要
					knownDeviceList[serviceId].measurCycSet = false;
					knownDeviceList[serviceId].timingStamp = 0;
				}else{
					let timeStamp           = utl.isoDateString2milliSeconds(timeStampString);
					let uvi                 = getUvI(retDat);
					let pm10                = getPm10(retDat);
					let pm25                = getPm25(retDat);
					let temperature         = getTemperature(retDat);
					let humidity            = getHumidity(retDat);
					let illuminance         = getIlluminance(retDat);
					let tvoc                = getTvoc(retDat);
					let atmosphericPressure = getAtmosphericPressure(retDat);
					let co2                 = getECo2(retDat);

					utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, "(a)bleId               :" + bleId);
					utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, " (a)timeStampString    :" + timeStampString);
					utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, " (a)timeStamp          :" + timeStamp);
					utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, " (a)uv                 :" + uvi);
					utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, " (a)pm10               :" + pm10);
					utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, " (a)pm25               :" + pm25);
					utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, " (a)temperature        :" + temperature);
					utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, " (a)humidity           :" + humidity);
					utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, " (a)illuminance        :" + illuminance);
					utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, " (a)voc                :" + tvoc);
					utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, " (a)atmosphericPressure:" + atmosphericPressure);
					utl.debugLog(DEBG_LV1, DEBUG_LEVEL, DEVICE_KIND, " (a)co2                :" + co2);

					// 測定値時間を記録(デバイスの時間)
					knownDeviceList[serviceId].timeStamp       = timeStamp;
					knownDeviceList[serviceId].timeStampString = timeStampString;
					knownDeviceList[serviceId].timingStamp     = timeStamp;

					// New測定値を記録
					knownDeviceList[serviceId].atmosphericPressure.value = atmosphericPressure;
					knownDeviceList[serviceId].humidity.value = humidity;
					knownDeviceList[serviceId].illuminance.value = illuminance;
					knownDeviceList[serviceId].pm10.value = pm10;
					knownDeviceList[serviceId].pm25.value = pm25;
					knownDeviceList[serviceId].temperature.value = temperature;
					knownDeviceList[serviceId].uv.value = uvi;
					knownDeviceList[serviceId].voc.value = tvoc;
					knownDeviceList[serviceId].co2Concentration.value = co2;
				}

				// TimeOut Clear
				_clearEventTimer(bleId);

				utl.debugLog(1, DEBUG_LEVEL, DEVICE_KIND, "notify disable実施");
				await nodeBleLib.stopNotifications(characteristic, BLE_LIB_TIMEOUT);

				resolve(true);
			});

			// notifyの受け場所を enable
			utl.debugLog(1, DEBUG_LEVEL, DEVICE_KIND, "notify enable実施");
			let retStartNoti = await nodeBleLib.startNotifications(characteristic, retryLimit, errorInterval, BLE_LIB_TIMEOUT);
			if(retStartNoti.result === false){
				utl.debugLog(1, DEBUG_LEVEL, DEVICE_KIND, "[error]nodeBleLib.startNotifications(" + CHARACTERISTIC_MSDP_NOTIFY + ")=" + retStartNoti.e);
				resolve(false);
				return;
			}

			// ガードタイマー設置
			_clearEventTimer(bleId);
			_startEventTimer(bleId, WAIT_TIME_OUT_API, async function(){
				await nodeBleLib.stopNotifications(characteristic, BLE_LIB_TIMEOUT);
				resolve(false);
			});

			// Lastデータ['F0']の取得実施
			let retLstDat;
			const buf = Buffer.from(READ_LAST_DATA, 'hex');
			flgLimit = true;
			utl.debugLog(1, DEBUG_LEVEL, DEVICE_KIND, "Lastデータ['F0']取得実施");
			for(let i=0; i<retryLimit; i++){
				retLstDat = await nodeBleLib.writeCharacteristic(characteristic, buf, "command", WAIT_TIME_OUT_API);
				if(retLstDat.result === true){
					flgLimit = false;
					break;
				}else{
					await utl.waitFunc(errorInterval);
				}
			}
			if(flgLimit){
				utl.debugLog(1, DEBUG_LEVEL, DEVICE_KIND, "[error]nodeBleLib.writeCharacteristic(" + CHARACTERISTIC_LATEST_MEASUREMENT + ")=" + buf.toString());
				await bleLib.stopNotifications(characteristic, BLE_LIB_TIMEOUT);
				resolve(false);
				return;
			}
		}catch(e){
			utl.debugLog(1, DEBUG_LEVEL, DEVICE_KIND, "[error]_getAutoMeasuredData(" + CHARACTERISTIC_LATEST_MEASUREMENT + ")=" + e);
			resolve(false);
			return;
		}
	});
};



// センサーデータ取得
// 計測データ変換
function convertMeasurementData(buffer, size){
	let retAryTmp = [size];
	for(let i=0; i<size; i++){
		// 受けたものを配列に突っ込む
		retAryTmp[i] = parseInt(buffer.toString('hex', i, i+1), 16);
	}
	return retAryTmp;
};
// 日時(計測時 uint16_t 2)
function getDateTime(dataArr){
	// Bit 4～0: 常に0
	// Bit 10～5: 0～63 (0～59 分)
	// Bit 15～11: 0～31 (0～23 時)
	let byte1Tim = dataArr[0];
	let byte2Tim = dataArr[1];
	let timeDat  = ((byte2Tim << 8) | byte1Tim);
	timeDat = (timeDat >> 5);
	let minDat = (timeDat & 0x3f);
	let hurDat = ((timeDat >> 6) & 0x1f);

	// Bit 4～0: 0～31 (1～31 日)
	// Bit 8～5: 0～15 (1～12 月)
	// Bit 15～9: 0～127 (00～99 年) 
	byte1Tim = dataArr[2];
	byte2Tim = dataArr[3];
	let dateDat = ((byte2Tim << 8) | byte1Tim);
	let dayDat = (byte1Tim & 0x1f);
	let mthDat = ((dateDat >> 5) & 0xf);
	let yerDat = ((byte2Tim >> 1) & 0x7f) + 2000;

	let dateString = yerDat + "-" + utl.zeroPadding2Dec(mthDat, 2) + "-" + utl.zeroPadding2Dec(dayDat, 2) + "T" + utl.zeroPadding2Dec(hurDat, 2) + ":" + utl.zeroPadding2Dec(minDat, 2) + ":00+09:00";
	return dateString;
};
// PM2.5濃度[μg/㎥]
function getPm25(dataArr){
	// Bit 9～0: 0～1022 (25～1000[μg/㎥])
	// 1023:Err
	let byte1Tim = dataArr[4];
	let byte2Tim = dataArr[5];
	let pm25eDat  = ((byte2Tim << 8) | byte1Tim) & 0x3ff;

	if(pm25eDat === 0x3ff){
		// 1023 = Error
		return -1;
	}else{
		pm25eDat = Math.round(pm25eDat*10)/10;
		return pm25eDat;
	}
};
// PM1.0濃度[μg/㎥]
function getPm10(dataArr){
	// Bit 9～0: 0～1022 (25～1000[μg/㎥])
	// 1023:Err
	let byte1Tim = dataArr[5];
	let byte2Tim = dataArr[6];
	let pm10Dat  = ((byte2Tim << 8) | byte1Tim);
	pm10Dat = (pm10Dat >> 3) & 0x3ff;

	if(pm10Dat === 0x3ff){
		// 1023 = Error
		return -1;
	}else{
		pm10Dat = Math.round(pm10Dat*10)/10;
		return pm10Dat;
	}
};
// UVインデックス
function getUvI(dataArr){
	// Bit 23～20: 0～14 (0～11[UVI])
	// 15:Erro
	let byte1Tim = dataArr[6];
	let uviDat = ((byte1Tim >> 4) | byte1Tim) & 0x0f;

	if(uviDat === 15){
		// 15 = Error
		return -9999;
	}else{
		return uviDat;
	}
};
// 温度[℃]
function getTemperature(dataArr){
	// Bit 10～0: 0～2046  【÷10-40.0】 =-40.0～164.6(-20.0～85.0[℃]) 
	// 2047:Error 
	let byte1Tim = dataArr[8];
	let byte2Tim = dataArr[9];
	let tmprDat  = ((byte2Tim << 8) | byte1Tim) & 0x7ff;
	if(tmprDat === 2047){
		// 2047 = Error
		return -9999;
	}else{
		tmprDat = tmprDat/10;
		tmprDat = tmprDat-40;
		tmprDat = Math.round(tmprDat*10)/10;
		return tmprDat;
	}
};
// 湿度[%/RH]
function getHumidity(dataArr){
	// Bit 20～11: 0～1022 【÷10】  = 0.0～102.2 (0.0～100.0[%/RH])
	// 1023:Error 
	let byte1Tim = dataArr[9];
	let byte2Tim = dataArr[10];
	let humiDat  = ((byte2Tim << 8) | byte1Tim);
	humiDat = (humiDat >> 3) & 0x3ff;
	if(humiDat === 1023){
		// 1023 = Error
		return -9999;
	}else{
		humiDat = humiDat/10;
		humiDat = Math.round(humiDat*10)/10;
		return humiDat;
	}
};
// 気圧[hPa]
function getAtmosphericPressure(dataArr){
	// (0 ～16382)/10 = (300.0～1100.0[hPa])
	// 16383:Error 
	let byte1Tim = dataArr[12];
	let byte2Tim = dataArr[13];
	let atmoDat  = ((byte2Tim << 8) | byte1Tim);
	atmoDat = atmoDat & 0x3fff;
	if(atmoDat === 16383){
		// 16383 = Error
		return -9999;
	}else{
		atmoDat = atmoDat/10;
		atmoDat = Math.round(atmoDat*10)/10;
		return atmoDat;
	}
};
// 照度[lx]
function getIlluminance(dataArr){
	// 0 ～ 65534  (0～65534[lx])
	// 65535:Error 
	let byte1Tim = dataArr[14];
	let byte2Tim = dataArr[15];
	let illDat  = ((byte2Tim << 8) | byte1Tim);
	if(illDat === 65535){
		// 65535 = Error
		return -9999;
	}else{
		return illDat;
	}
};
// Tvoc[ppb]
function getTvoc(dataArr){
	// 0～2046  (0～1187[ppb])
	// 2047:Error 
	let byte1Tim = dataArr[16];
	let byte2Tim = dataArr[17];
	let vocDat  = ((byte2Tim << 8) | byte1Tim);
	vocDat = vocDat & 0x07ff;
	if(vocDat === 2047){
		// 2047 = Error
		return -9999;
	}else{
		return vocDat;
	}
};
// eCo2[ppm]
function getECo2(dataArr){
	// 0～8190  (400～8190[ppm])
	// 8191:Error 
	let byte1Tim = dataArr[17];
	let byte2Tim = dataArr[18];
	let co2Dat  = ((byte2Tim << 8) | byte1Tim);
	co2Dat = ((co2Dat >> 3) & 0x1FFF);
	if(co2Dat === 8191){
		// 8191 = Error
		return -9999;
	}else{
		return co2Dat;
	}
};
// VOCセンサー状態：起動
function getVocOn(bit32Arr){
	return parseInt(bit32Arr[3].slice(16, 17), 2);
};
// VOCセンサー状態：初期
function getVocDefault(bit32Arr){
	return parseInt(bit32Arr[3].slice(17, 18), 2);
};
// 計測データモード
function getMeasurementMode(bit32Arr){
	return parseInt(bit32Arr[4].slice(0, 8), 2);
};
// 時刻から16進数時刻データを取得する
const getHexDatetime = (dt) => {
	const seconds = ('0' + dt.getSeconds().toString(16)).slice(-2);
	const minutes = ('0' + dt.getMinutes().toString(16)).slice(-2);
	const hours   = ('0' + dt.getHours().toString(16)).slice(-2);
	const date    = ('0' + dt.getDate().toString(16)).slice(-2);
	const month   = ('0' + (dt.getMonth() + 1).toString(16)).slice(-2);
	const year    = Number(dt.getFullYear().toString().slice(-2)).toString(16);
	const week    = ('0' + dt.getDay());
	return `${seconds}${minutes}${hours}${date}${month}${year}${week}`;
};
// unix時刻をYYYYMMDDHHMMSS.sss +/- HHMMに変換
const unixtimeToDatetime = (ut) => {
	const dt = new Date(ut);
	const year = dt.getFullYear();
	const month = ('0' + (dt.getMonth() + 1)).slice(-2);
	const date = ('0' + dt.getDate()).slice(-2);
	const hours = ('0' + dt.getHours()).slice(-2);
	const minutes = ('0' + dt.getMinutes()).slice(-2);
	const seconds = ('0' + dt.getSeconds()).slice(-2);
	const mseconds = ut.toString().slice(-3);
	return `${year}${month}${date}${hours}${minutes}${seconds}.${mseconds}+0900`;
};

async function testStart(){
	// 初期テーブル作成
	await deviceInitReg();

	// 受信開始
	setTimeout(() => {
		_startDeviceFinder();
	}, START_DELAY_TIME);
};

testStart();
